from typing import List, Iterator


class HomiException(Exception):
    pass


class RegisterError(HomiException):
    pass


class ProtobufError(HomiException):
    pass


class ServiceNotFound(ProtobufError):
    def __init__(self, service_name, available_services: Iterator[str] = None, **kwargs):
        self.fail_service = service_name
        self.services = available_services

    def __str__(self):
        msg = f"Can not find {self.fail_service} in app. you must add service"
        return f"{msg} or you can use this services {', '.join(self.services)}" if self.services else self.msg


class MethodNotFound(ProtobufError):
    def __init__(self, method_name, service_name=None, available_methods: List[str] = None, **kwargs):
        self.fail_method = method_name
        self.service_name = service_name or ''
        self.methods = available_methods

    def __str__(self):
        msg = f"Can not find {self.fail_method} in {self.service_name}."
        return f"{msg} available methods is {', '.join(self.methods)}" if self.methods else self.msg



from concurrent import futures

import grpc

from .app import App


class Server:
    def __init__(self, app: App, host: str = '127.0.0.1', port: str = '50051', worker: int = 10):
        self.host = host
        self.port = port
        self.worker = worker
        self.app = app
        self.load_config_from_env()
        self.server = None
        self.thread_pool = futures.ThreadPoolExecutor(max_workers=self.worker)

    def load_config_from_env(self):
        pass

    def run(self, wait=True):
        self.server = grpc.server(self.thread_pool)
        self.app.bind_to_server(self.server)

        self.server.add_insecure_port(f'[::]:{self.port}')
        self.server.start()
        print('run server')
        print(f'# port : {self.port}')
        if wait:
            self.wait_for_termination()

    def stop(self, grace=None):
        if self.server:
            self.server.stop(grace)

    def wait_for_termination(self):
        if self.server:
            self.server.wait_for_termination()



import unittest
from typing import Any, Dict

import grpc_testing

from . import Server, App, Service


class HomiTestCase(unittest.TestCase):
    app: App = None
    _test_server = None

    def get_test_server(self):
        if not self._test_server:
            servicers = {}
            for svc in self.app.services:
                if isinstance(svc, Service):
                    servicers[svc.descriptor] = svc.make_servicer_class()

            self._test_server = grpc_testing.server_from_dictionary(
                servicers, grpc_testing.strict_real_time()
            )
        return self._test_server

    @staticmethod
    def get_all_response(method):
        finish = False
        result = []
        while not finish:
            try:
                result.append(method.take_response())
            except Exception:
                finish = True
        return result

    @staticmethod
    def send_request_all(method, requests):
        [method.send_request(req) for req in requests]
        method.requests_closed()


class HomiRealServerTestCase(unittest.TestCase):
    app = None
    default_server_config = {
        "host": "localhost",
        "port": '5999'
    }
    test_server_config: Dict[str, Any] = {}
    test_server = None

    def get_server_config(self, merge_config: dict = None):
        config = merge_config or {}
        return {
            **self.default_server_config,
            **self.test_server_config,
            **config,
        }

    def server_restart(self, merge_config: dict = None):
        self.run_real_server(merge_config)

    def run_real_server(self, merge_config: dict = None):
        config = merge_config or {}
        if self.test_server:
            try:
                self.test_server.stop()
            except Exception:
                pass
        self.test_server = Server(**self.get_server_config(config))
        self.test_server.run(wait=False)

    def setUp(self):
        self.run_real_server()

    def tearDown(self):
        self.test_server.stop()
